"""Code for loading and manipulating data structures."""
from .base import CoordSpec, DimSpec, dict_to_dataset, numpy_to_data_array
from .converters import convert_to_dataset, convert_to_inference_data
from .datasets import clear_data_home, list_datasets, load_arviz_data
from .inference_data import InferenceData, concat
from .io_beanmachine import from_beanmachine
from .io_cmdstan import from_cmdstan
from .io_cmdstanpy import from_cmdstanpy
from .io_datatree import from_datatree, to_datatree
from .io_dict import from_dict
from .io_emcee import from_emcee
from .io_json import from_json, to_json
from .io_netcdf import from_netcdf, to_netcdf
from .io_numpyro import from_numpyro
from .io_pyjags import from_pyjags
from .io_pyro import from_pyro
from .io_pystan import from_pystan
from .io_zarr import from_zarr, to_zarr
from .utils import extract, extract_dataset

__all__ = [
    "InferenceData",
    "concat",
    "load_arviz_data",
    "list_datasets",
    "clear_data_home",
    "numpy_to_data_array",
    "extract",
    "extract_dataset",
    "dict_to_dataset",
    "convert_to_dataset",
    "convert_to_inference_data",
    "from_beanmachine",
    "from_pyjags",
    "from_pystan",
    "from_emcee",
    "from_cmdstan",
    "from_cmdstanpy",
    "from_datatree",
    "from_dict",
    "from_json",
    "from_pyro",
    "from_numpyro",
    "from_netcdf",
    "to_datatree",
    "to_json",
    "to_netcdf",
    "from_zarr",
    "to_zarr",
    "CoordSpec",
    "DimSpec",
]
